from .generate_detachment_ltd_erosion import DetachmentLtdErosion
from .generate_erosion_by_depth_slope import DepthSlopeProductErosion

__all__ = ['DetachmentLtdErosion',  'DepthSlopeProductErosion']